#   Copyright 2020-2020 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


from th2_common.schema.message.configuration.queue_configuration import QueueConfiguration


class MessageRouterConfiguration:
    def __init__(self, queues: dict) -> None:
        self.queues = {queue_alias: QueueConfiguration(**queues[queue_alias]) for queue_alias in queues.keys()}

    def get_queue_by_alias(self, queue_alias):
        return self.queues[queue_alias]

    def find_queues_by_attr(self, attrs) -> {str: QueueConfiguration}:
        result = dict()
        for queue_alias in self.queues.keys():
            ok = True
            for attr in attrs:
                if not self.queues[queue_alias].attributes.__contains__(attr):
                    ok = False
                    break
            if ok:
                result[queue_alias] = self.queues[queue_alias]
        return result



#   Copyright 2020-2021 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from th2_common.schema.filter.strategy.impl.default_filter_strategy import DefaultFilterStrategy
from th2_common.schema.message.configuration.queue_configuration import QueueConfiguration
from th2_common.schema.message.impl.rabbitmq.abstract_rabbit_queue import AbstractRabbitQueue
from th2_common.schema.message.impl.rabbitmq.configuration.subscribe_target import SubscribeTarget
from th2_common.schema.message.impl.rabbitmq.connection.connection_manager import ConnectionManager
from th2_common.schema.message.impl.rabbitmq.parsed.rabbit_parsed_batch_sender import RabbitParsedBatchSender
from th2_common.schema.message.impl.rabbitmq.parsed.rabbit_parsed_batch_subscriber import RabbitParsedBatchSubscriber
from th2_common.schema.message.message_sender import MessageSender
from th2_common.schema.message.message_subscriber import MessageSubscriber


class RabbitParsedBatchQueue(AbstractRabbitQueue):

    def create_sender(self, connection,
                      queue_configuration: QueueConfiguration) -> MessageSender:
        return RabbitParsedBatchSender(connection, queue_configuration.exchange, queue_configuration.name)

    def create_subscriber(self, connection_manager: ConnectionManager,
                          queue_configuration: QueueConfiguration) -> MessageSubscriber:
        subscribe_target = SubscribeTarget(routing_key=queue_configuration.name, queue=queue_configuration.queue)
        return RabbitParsedBatchSubscriber(connection_manager,
                                           queue_configuration,
                                           DefaultFilterStrategy(),
                                           subscribe_target)



#   Copyright 2020-2020 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


from abc import ABC
from concurrent.futures.thread import ThreadPoolExecutor

import grpc

from th2_common.schema.grpc.configuration.grpc_router_configuration import GrpcRouterConfiguration
from th2_common.schema.grpc.router.grpc_router import GrpcRouter


class AbstractGrpcRouter(GrpcRouter, ABC):

    def __init__(self, configuration: GrpcRouterConfiguration) -> None:
        self.configuration = configuration
        self.servers = []
        self.channels = {}

    def start_server(self, *services) -> grpc.Server:
        server = grpc.server(ThreadPoolExecutor(max_workers=self.configuration.serverConfiguration.workers))

        if self.configuration.serverConfiguration.host is None:
            server.add_insecure_port(f'[::]:{self.configuration.serverConfiguration.port}')
        else:
            server.add_insecure_port(
                f'{self.configuration.serverConfiguration.host}:{self.configuration.serverConfiguration.port}')

        self.servers.append(server)

        return server

    def close(self, grace=None):
        for server in self.servers:
            server.stop(grace)

        for channel in self.channels.values():
            channel.close()



#   Copyright 2020-2021 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import datetime
import logging
import time
from abc import ABC, abstractmethod
from threading import Lock

from google.protobuf.message import DecodeError
from prometheus_client import Histogram, Counter

from th2_common.schema.exception.router_error import RouterError
from th2_common.schema.message.configuration.queue_configuration import QueueConfiguration
from th2_common.schema.message.impl.rabbitmq.connection.connection_manager import ConnectionManager
from th2_common.schema.message.message_listener import MessageListener
from th2_common.schema.message.message_subscriber import MessageSubscriber

logger = logging.getLogger()


class AbstractRabbitSubscriber(MessageSubscriber, ABC):

    def __init__(self, connection_manager: ConnectionManager, queue_configuration: QueueConfiguration,
                 *subscribe_targets) -> None:
        if len(subscribe_targets) < 1:
            raise Exception('Subscribe targets must be more than 0')

        self.listeners = set()
        self.lock_listeners = Lock()

        self.connection_manager = connection_manager
        self.channel = None

        self.subscribe_targets = subscribe_targets
        self.subscriber_name = connection_manager.configuration.subscriber_name

        self.prefetch_count = queue_configuration.prefetch_count
        self.exchange_name = queue_configuration.exchange
        self.attributes = tuple(set(queue_configuration.attributes))

    def start(self):
        if self.subscribe_targets is None or self.exchange_name is None:
            raise Exception('Subscriber did not init')

        if self.subscriber_name is None:
            self.subscriber_name = 'rabbit_mq_subscriber'
            logger.info(f"Using default subscriber name: '{self.subscriber_name}'")

        if self.channel is None:
            self.channel = self.connection_manager.connection.channel()
            channel_open_timeout = 50
            for x in range(int(channel_open_timeout / 5)):
                if not self.channel.is_open:
                    time.sleep(5)
            if not self.channel.is_open:
                raise RouterError(f"The channel has not been opened for {channel_open_timeout} seconds")
            logger.info(f"Open channel: {self.channel} for subscriber[{self.exchange_name}]")

            for subscribe_target in self.subscribe_targets:
                queue = subscribe_target.get_queue()
                routing_key = subscribe_target.get_routing_key()
                self.channel.basic_qos(prefetch_count=self.prefetch_count)
                consumer_tag = f'{self.subscriber_name}.{datetime.datetime.now()}'
                self.channel.basic_consume(queue=queue, consumer_tag=consumer_tag,
                                           on_message_callback=self.handle)

                logger.info(f"Start listening exchangeName='{self.exchange_name}', "
                            f"routing key='{routing_key}', queue name='{queue}', consumer_tag={consumer_tag}")

    def is_close(self) -> bool:
        return self.channel is None or not self.channel.is_open

    def close(self):
        with self.lock_listeners:
            for listener in self.listeners:
                listener.on_close()
            self.listeners.clear()

        if self.channel is not None and self.channel.is_open:
            self.channel.close()

    def add_listener(self, message_listener: MessageListener):
        if message_listener is None:
            return
        with self.lock_listeners:
            self.listeners.add(message_listener)

    @abstractmethod
    def get_delivery_counter(self) -> Counter:
        pass

    @abstractmethod
    def get_content_counter(self) -> Counter:
        pass

    @abstractmethod
    def get_processing_timer(self) -> Histogram:
        pass

    @abstractmethod
    def extract_count_from(self, message):
        pass

    def handle(self, channel, method, properties, body):
        try:
            process_timer = self.get_processing_timer()
            start_time = time.time()

            value = self.value_from_bytes(body)

            if value is None:
                raise ValueError('Received value is null')

            counter = self.get_delivery_counter()
            counter.inc()
            content_counter = self.get_content_counter()
            content_counter.inc(self.extract_count_from(value))

            if not self.filter(value):
                return

            self.handle_with_listener(value, channel, method)

            end_time = time.time()
            process_timer.observe(end_time - start_time)

        except DecodeError as e:
            logger.exception(
                f'Can not parse value from delivery for: {method.consumer_tag} due to DecodeError: {e}\n'
                f'  body: {body}\n'
                f'  self: {self}\n')
            return
        except Exception as e:
            logger.error(f'Can not parse value from delivery for: {method.consumer_tag}', e)
            return
        finally:
            if channel.is_open:
                channel.basic_ack(method.delivery_tag)
            else:
                logger.error('Message acknowledgment failed due to the channel being closed')

    def handle_with_listener(self, value, channel, method):
        with self.lock_listeners:
            for listener in self.listeners:
                try:
                    listener.handler(self.attributes, value)
                except Exception as e:
                    logger.warning(f"Message listener from class '{type(listener)}' threw exception {e}")

    @abstractmethod
    def value_from_bytes(self, body):
        pass

    @abstractmethod
    def filter(self, value) -> bool:
        pass



#   Copyright 2020-2021 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from prometheus_client import Counter, Histogram
from th2_grpc_common.common_pb2 import RawMessageBatch, RawMessage

from th2_common.schema.message.impl.rabbitmq.abstract_rabbit_batch_subscriber import AbstractRabbitBatchSubscriber, \
    Metadata
from th2_common.schema.metrics.common_metrics import CommonMetrics


class RabbitRawBatchSubscriber(AbstractRabbitBatchSubscriber):
    INCOMING_RAW_MSG_BATCH_QUANTITY = Counter('th2_mq_incoming_raw_msg_batch_quantity',
                                              'Quantity of incoming raw message batches')
    INCOMING_RAW_MSG_QUANTITY = Counter('th2_mq_incoming_raw_msg_quantity',
                                        'Quantity of incoming raw messages')
    RAW_MSG_PROCESSING_TIME = Histogram('th2_mq_raw_msg_processing_time',
                                        'Time of processing raw messages',
                                        buckets=CommonMetrics.DEFAULT_BUCKETS)

    __MESSAGE_TYPE = 'raw'

    def get_delivery_counter(self) -> Counter:
        return self.INCOMING_RAW_MSG_BATCH_QUANTITY

    def get_content_counter(self) -> Counter:
        return self.INCOMING_RAW_MSG_QUANTITY

    def get_processing_timer(self) -> Histogram:
        return self.RAW_MSG_PROCESSING_TIME

    def extract_count_from(self, message: RawMessageBatch):
        return len(self.get_messages(message))

    def get_messages(self, batch: RawMessageBatch) -> list:
        return batch.messages

    def value_from_bytes(self, body):
        message_batch = RawMessageBatch()
        message_batch.ParseFromString(body)
        return message_batch

    def extract_metadata(self, message: RawMessage) -> Metadata:
        metadata = message.metadata
        return Metadata(message_type=self.__MESSAGE_TYPE,
                        direction=metadata.id.direction,
                        sequence=metadata.id.sequence,
                        session_alias=metadata.id.connection_id.session_alias)



#   Copyright 2020-2020 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


from abc import ABC, abstractmethod

from google.protobuf.message import Message

from th2_common.schema.message.configuration.queue_configuration import QueueConfiguration
from th2_common.schema.message.impl.rabbitmq.abstract_rabbit_message_router import AbstractRabbitMessageRouter


class AbstractRabbitBatchMessageRouter(AbstractRabbitMessageRouter, ABC):

    def _find_by_filter(self, queues: {str: QueueConfiguration}, batch) -> dict:
        result = dict()
        for message in self._get_messages(batch):
            for queue_alias in self._filter(queues, message):
                self._add_message(result.setdefault(queue_alias, self._create_batch()), message)
        return result

    def _filter(self, queues: {str: QueueConfiguration}, message: Message) -> {str}:
        aliases = set()
        for queue_alias in queues.keys():
            filters = queues[queue_alias].filters

            if len(filters) == 0 or self._filter_strategy.verify(message, filters):
                aliases.add(queue_alias)
        return aliases

    @abstractmethod
    def _get_messages(self, batch) -> list:
        pass

    @abstractmethod
    def _create_batch(self):
        pass

    @abstractmethod
    def _add_message(self, batch, message):
        pass



#   Copyright 2020-2021 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
import time
from abc import ABC, abstractmethod
from threading import Lock

from pika.exceptions import ConnectionWrongStateError
from prometheus_client import Counter

from th2_common.schema.exception.router_error import RouterError
from th2_common.schema.message.impl.rabbitmq.connection.connection_manager import ConnectionManager
from th2_common.schema.message.message_sender import MessageSender

logger = logging.getLogger()


class AbstractRabbitSender(MessageSender, ABC):

    def __init__(self, connection_manager: ConnectionManager, exchange_name: str, send_queue: str) -> None:
        self.connection_manager = connection_manager
        self.channel = None
        self.exchange_name = exchange_name
        self.send_queue = send_queue
        self.__lock = Lock()

    def start(self):
        if self.send_queue is None or self.exchange_name is None:
            raise Exception('Sender can not start. Sender did not init')
        if self.channel is None:
            self.channel = self.connection_manager.connection.channel()
            channel_open_timeout = 60
            for x in range(int(channel_open_timeout / 5)):
                if not self.channel.is_open:
                    time.sleep(5)
            if not self.channel.is_open:
                raise RouterError(f"The channel has not been opened for {channel_open_timeout} seconds")
            logger.info(f"Open channel: {self.channel} for sender[{self.exchange_name}:{self.send_queue}]")

    def is_close(self) -> bool:
        return self.channel is None or not self.channel.is_open

    def close(self):
        if self.channel is not None and self.channel.is_open:
            logger.info(f"Close channel: {self.channel} for sender[{self.exchange_name}:{self.send_queue}]")
            self.channel.close()

    @abstractmethod
    def get_delivery_counter(self) -> Counter:
        pass

    @abstractmethod
    def get_content_counter(self) -> Counter:
        pass

    @abstractmethod
    def extract_count_from(self, message):
        pass

    def send(self, message):
        if message is None:
            raise ValueError('Value for send can not be null')

        with self.__lock:
            try:
                for x in self.connection_manager.reconnect_attempts:
                    try:
                        self.channel.basic_publish(exchange=self.exchange_name, routing_key=self.send_queue,
                                                   body=self.value_to_bytes(message))

                        counter = self.get_delivery_counter()
                        counter.inc()
                        content_counter = self.get_content_counter()
                        content_counter.inc(self.extract_count_from(message))

                        logger.info(f"Sent message:\n{message}"
                                    f"Exchange: '{self.exchange_name}', routing key: '{self.send_queue}'")
                        break
                    except ConnectionWrongStateError:
                        self.connection_manager.reopen_connection()
            except Exception:
                if self.channel is None:
                    raise Exception('Can not send. Sender did not started')
                raise

    @abstractmethod
    def value_to_bytes(self, value):
        pass



#   Copyright 2020-2021 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from prometheus_client import Gauge


class CommonMetrics:
    DEFAULT_BUCKETS = [0.000_25, 0.000_5, 0.001, 0.005, 0.010, 0.015, 0.025, 0.050, 0.100, 0.250, 0.500, 1.0]

    LIVENESS = Gauge("th2_liveness", "Service liveness")
    READINESS = Gauge("th2_readiness", "Service readiness")

    RABBITMQ_READINESS = True
    GRPC_READINESS = True
    ALL_READINESS = [RABBITMQ_READINESS, GRPC_READINESS]

    def check_common_readiness(self):
        for readiness_parameter in self.ALL_READINESS:
            if not readiness_parameter:
                return False
        return True



#   Copyright 2020-2020 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


import argparse
import logging
import sys
from json import loads, dumps
from os import mkdir, getcwd

from kubernetes import client, config

from th2_common.schema.event.event_batch_router import EventBatchRouter
from th2_common.schema.factory.abstract_common_factory import AbstractCommonFactory
from th2_common.schema.grpc.router.impl.default_grpc_router import DefaultGrpcRouter
from th2_common.schema.message.impl.rabbitmq.parsed.rabbit_parsed_batch_router import RabbitParsedBatchRouter
from th2_common.schema.message.impl.rabbitmq.raw.rabbit_raw_batch_router import RabbitRawBatchRouter

logger = logging.getLogger()


class CommonFactory(AbstractCommonFactory):
    CONFIG_DEFAULT_PATH = '/var/th2/config/'

    RABBIT_MQ_CONFIG_FILENAME = 'rabbitMQ.json'
    MQ_ROUTER_CONFIG_FILENAME = 'mq.json'
    GRPC_ROUTER_CONFIG_FILENAME = 'grpc.json'
    CRADLE_CONFIG_FILENAME = 'cradle.json'
    PROMETHEUS_CONFIG_FILENAME = 'prometheus.json'
    CUSTOM_CONFIG_FILENAME = 'custom.json'

    #   FIX: Add path to dictionary as a parameter
    def __init__(self,
                 config_path=None,
                 rabbit_mq_config_filepath=CONFIG_DEFAULT_PATH + RABBIT_MQ_CONFIG_FILENAME,
                 mq_router_config_filepath=CONFIG_DEFAULT_PATH + MQ_ROUTER_CONFIG_FILENAME,
                 grpc_router_config_filepath=CONFIG_DEFAULT_PATH + GRPC_ROUTER_CONFIG_FILENAME,
                 cradle_config_filepath=CONFIG_DEFAULT_PATH + CRADLE_CONFIG_FILENAME,
                 prometheus_config_filepath=CONFIG_DEFAULT_PATH + PROMETHEUS_CONFIG_FILENAME,
                 custom_config_filepath=CONFIG_DEFAULT_PATH + CUSTOM_CONFIG_FILENAME,

                 message_parsed_batch_router_class=RabbitParsedBatchRouter,
                 message_raw_batch_router_class=RabbitRawBatchRouter,
                 event_batch_router_class=EventBatchRouter,
                 grpc_router_class=DefaultGrpcRouter) -> None:

        if config_path is not None:
            CommonFactory.CONFIG_DEFAULT_PATH = config_path
            rabbit_mq_config_filepath = config_path + CommonFactory.RABBIT_MQ_CONFIG_FILENAME
            mq_router_config_filepath = config_path + CommonFactory.MQ_ROUTER_CONFIG_FILENAME
            grpc_router_config_filepath = config_path + CommonFactory.GRPC_ROUTER_CONFIG_FILENAME
            cradle_config_filepath = config_path + CommonFactory.CRADLE_CONFIG_FILENAME
            prometheus_config_filepath = config_path + CommonFactory.PROMETHEUS_CONFIG_FILENAME
            custom_config_filepath = config_path + CommonFactory.CUSTOM_CONFIG_FILENAME

        self.rabbit_mq_config_filepath = rabbit_mq_config_filepath
        self.mq_router_config_filepath = mq_router_config_filepath
        self.grpc_router_config_filepath = grpc_router_config_filepath
        self.cradle_config_filepath = cradle_config_filepath
        self.prometheus_config_filepath = prometheus_config_filepath
        self.custom_config_filepath = custom_config_filepath

        super().__init__(message_parsed_batch_router_class, message_raw_batch_router_class,
                         event_batch_router_class, grpc_router_class)

    @staticmethod
    def calculate_path(parsed_args, name_attr, path_default) -> str:
        if parsed_args.__contains__(name_attr):
            return parsed_args.__getattribute__(name_attr)
        else:
            return CommonFactory.CONFIG_DEFAULT_PATH + path_default

    @staticmethod
    def create_from_arguments(args=None):
        if args is None:
            args = sys.argv[1:]

        parser = argparse.ArgumentParser()
        parser.add_argument('--rabbitConfiguration',
                            help='path to json file with RabbitMQ configuration')
        parser.add_argument('--messageRouterConfiguration',
                            help='path to json file with configuration for MessageRouter')
        parser.add_argument('--grpcRouterConfiguration',
                            help='path to json file with configuration for GrpcRouter')
        parser.add_argument('--cradleConfiguration',
                            help='path to json file with configuration for cradle')
        parser.add_argument('--prometheusConfiguration',
                            help='path to json file with configuration for prometheus metrics server')
        parser.add_argument('--namespace',
                            help='namespace in Kubernetes to find config maps related to the target')
        parser.add_argument('--boxName',
                            help='name of the target th2 box placed in the specified namespace in Kubernetes')
        parser.add_argument('--customConfiguration',
                            help='path to json file with custom configuration')
        result = parser.parse_args(args)

        if hasattr(result, 'namespace') and hasattr(result, 'box_name'):

            return CommonFactory.create_from_kubernetes(result.namespace, result.box_name)

        else:

            return CommonFactory(
                rabbit_mq_config_filepath=CommonFactory.calculate_path(result, 'rabbitConfiguration',
                                                                       CommonFactory.RABBIT_MQ_CONFIG_FILENAME),
                mq_router_config_filepath=CommonFactory.calculate_path(result, 'messageRouterConfiguration',
                                                                       CommonFactory.MQ_ROUTER_CONFIG_FILENAME),
                grpc_router_config_filepath=CommonFactory.calculate_path(result, 'grpcRouterConfiguration',
                                                                         CommonFactory.GRPC_ROUTER_CONFIG_FILENAME),
                cradle_config_filepath=CommonFactory.calculate_path(result, 'cradleConfiguration',
                                                                    CommonFactory.CRADLE_CONFIG_FILENAME),
                prometheus_config_filepath=CommonFactory.calculate_path(result, 'prometheusConfiguration',
                                                                        CommonFactory.PROMETHEUS_CONFIG_FILENAME),
                custom_config_filepath=CommonFactory.calculate_path(result, 'customConfiguration',
                                                                    CommonFactory.CUSTOM_CONFIG_FILENAME)
            )

    @staticmethod
    def create_from_kubernetes(namespace, box_name):

        config.load_kube_config()
        kube_contexts = config.kube_config.list_kube_config_contexts()

        v1 = client.CoreV1Api()

        config_maps = v1.list_namespaced_config_map(namespace)
        config_maps_dict = config_maps.to_dict()

        service_in_namespace_service = v1.list_namespaced_service('service').to_dict()

        services = v1.list_namespaced_service(namespace)
        services_dict = services.to_dict()

        config_dir = 'generated_configs'

        grpc_path = config_dir + '/' + CommonFactory.GRPC_ROUTER_CONFIG_FILENAME
        custom_path = config_dir + '/' + CommonFactory.CUSTOM_CONFIG_FILENAME
        mq_path = config_dir + '/' + CommonFactory.MQ_ROUTER_CONFIG_FILENAME
        cradle_path = config_dir + '/' + CommonFactory.CRADLE_CONFIG_FILENAME
        rabbit_path = config_dir + '/' + CommonFactory.RABBIT_MQ_CONFIG_FILENAME
        dictionary_path = config_dir + '/' + 'dictionary.json'
        prometheus_path = config_dir + '/' + 'prometheus.json'

        prometheus_host = '0.0.0.0'
        prometheus_port = 9752

        try:
            mkdir(config_dir)
            logger.info('Directory %s is created at %s' % (config_dir, getcwd()))
        except OSError:
            logger.info('All configuration in the %s folder are overridden' % (getcwd() + '/' + config_dir))

        CommonFactory._get_grpc(config_maps_dict, box_name, grpc_path, services_dict, kube_contexts)
        CommonFactory._get_custom(config_maps_dict, box_name, custom_path)
        CommonFactory._get_mq(config_maps_dict, box_name, mq_path)
        CommonFactory._get_cradle(config_maps_dict, cradle_path, kube_contexts, service_in_namespace_service)
        CommonFactory._get_rabbit(config_maps_dict, rabbit_path, kube_contexts, service_in_namespace_service)
        CommonFactory._get_dictionary(box_name, v1.list_config_map_for_all_namespaces(), dictionary_path)

        with open(prometheus_path, 'w') as prometheus_file:
            prometheus_file.write('{' + f'"host":"{prometheus_host}","port":{prometheus_port},"enabled":true' + '}')

        return CommonFactory(
            rabbit_mq_config_filepath=rabbit_path,
            mq_router_config_filepath=mq_path,
            grpc_router_config_filepath=grpc_path,
            cradle_config_filepath=cradle_path,
            prometheus_config_filepath=prometheus_path,
            custom_config_filepath=custom_path
        )

    @staticmethod
    def _get_dictionary(box_name, config_maps, dictionary_path):
        if 'items' in config_maps.to_dict()['items']:
            try:
                for config_map in config_maps.to_dict()['items']:
                    if config_map['metadata']['name'].startswith(box_name) & \
                            config_map['metadata']['name'].endswith('-dictionary'):
                        with open(dictionary_path, 'w') as dictionary_file:
                            dictionary_file.write(dumps(config_map))
            except KeyError:
                logger.error(f'dictionary config map\'s metadata not valid. Some keys are absent.')
            except IOError:
                logger.error('Failed to write file for dictionary.')

    @staticmethod
    def _get_grpc(config_maps_dict, box_name, grpc_path, services_dict, kube_contexts):
        try:
            if 'items' in config_maps_dict:
                for box_config_map in config_maps_dict['items']:
                    if box_config_map['metadata']['name'] == box_name + '-app-config':
                        box_data = box_config_map['data']
                        grpc = loads(box_data['grpc.json'])

                        for grpc_service in grpc['services']:
                            endpoints = grpc['services'][grpc_service]['endpoints']

                            for endpoint in endpoints:
                                endpoint_port = endpoints[endpoint]['port']

                                for service in services_dict['items']:
                                    if service['metadata']['name'] == endpoints[endpoint]['host']:
                                        ports = service['spec']['ports']

                                        for port in ports:
                                            if port['target_port'] == endpoint_port:
                                                grpc['services'][grpc_service]['endpoints'][endpoint]['port'] = \
                                                    port['node_port']
                                                grpc['services'][grpc_service]['endpoints'][endpoint]['host'] = \
                                                    kube_contexts[1]['name']

                                                with open(grpc_path, 'w') as grpc_file:
                                                    grpc_file.write(dumps(grpc))
                                                break
        except KeyError:
            logger.error(f'{box_name}-app-config\'s data not valid. Some keys are absent.')
        except IOError:
            logger.error('Failed to write grpc config.')

    @staticmethod
    def _get_custom(config_maps_dict, box_name, custom_path):
        try:
            if 'items' in config_maps_dict:
                for box_config_map in config_maps_dict['items']:
                    if box_config_map['metadata']['name'] == box_name + '-app-config':
                        box_data = box_config_map['data']
                        custom = loads(box_data['custom.json'])

                        with open(custom_path, 'w') as custom_file:
                            custom_file.write(dumps(custom))

        except KeyError:
            logger.error(f'{box_name}-app-config\'s data not valid. Some keys are absent.')
        except IOError:
            logger.error('Failed to write for custom config.')

    @staticmethod
    def _get_mq(config_maps_dict, box_name, mq_path):
        try:
            if 'items' in config_maps_dict:
                for box_config_map in config_maps_dict['items']:
                    if box_config_map['metadata']['name'] == box_name + '-app-config':
                        box_data = box_config_map['data']
                        mq = loads(box_data['mq.json'])

                        with open(mq_path, 'w') as mq_file:
                            mq_file.write(dumps(mq))

        except KeyError:
            logger.error(f'{box_name}-app-config\'s data not valid. Some keys are absent.')
        except IOError:
            logger.error('Failed to write for mq config.')

    @staticmethod
    def _get_cradle(config_maps_dict, cradle_path, kube_contexts, service_in_namespace_service):
        try:
            if 'items' in config_maps_dict:
                for cassandra_config_map in config_maps_dict['items']:
                    if cassandra_config_map['metadata']['name'] == 'cradle':
                        cradle = loads(cassandra_config_map['data']['cradle.json'])
                        cradle['host'] = kube_contexts[1]['name']

                        for item in service_in_namespace_service['items']:
                            if item['metadata']['name'] == 'cassandra-schema':

                                for port in item['spec']['ports']:
                                    if port['port'] == int(cradle['port']):
                                        cradle['port'] = port['node_port']

                                        with open(cradle_path, 'w') as cradle_file:
                                            cradle_file.write(dumps(cradle))
                        break
        except KeyError:
            logger.error(f'cassandra config map\'s data not valid. Some keys are absent.')
        except IOError:
            logger.error('Failed to write file for Cradle config.')

    @staticmethod
    def _get_rabbit(config_maps_dict, rabbit_path, kube_contexts, service_in_namespace_service):
        try:
            if 'items' in config_maps_dict:
                for rabbit_config_map in config_maps_dict['items']:
                    if rabbit_config_map['metadata']['name'] == 'rabbit-mq-app-config':
                        rabbit = loads(rabbit_config_map['data']['rabbitMQ.json'])
                        rabbit['host'] = kube_contexts[1]['name']
                        rabbit['username'] = kube_contexts[1]['context']['user']

                        for item in service_in_namespace_service['items']:
                            if item['metadata']['name'] == 'rabbitmq-schema':

                                for port in item['spec']['ports']:
                                    if port['port'] == int(rabbit['port']):
                                        rabbit['port'] = port['node_port']

                                        with open(rabbit_path, 'w') as rabbit_file:
                                            rabbit_file.write(dumps(rabbit))
                        break
        except KeyError:
            logger.error(f'rabbit config map\'s data not valid. Some keys are absent.')
        except IOError:
            logger.error(f'Failed to write file for rabbitMQ config.')

    def _path_to_rabbit_mq_configuration(self) -> str:
        return self.rabbit_mq_config_filepath

    def _path_to_message_router_configuration(self) -> str:
        return self.mq_router_config_filepath

    def _path_to_grpc_router_configuration(self) -> str:
        return self.grpc_router_config_filepath

    def _path_to_cradle_configuration(self) -> str:
        return self.cradle_config_filepath

    def _path_to_prometheus_configuration(self) -> str:
        return self.prometheus_config_filepath

    def _path_to_custom_configuration(self) -> str:
        return self.custom_config_filepath



#   Copyright 2020-2021 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from prometheus_client import Counter
from th2_grpc_common.common_pb2 import MessageBatch

from th2_common.schema.message.impl.rabbitmq.abstract_rabbit_sender import AbstractRabbitSender


class RabbitParsedBatchSender(AbstractRabbitSender):
    OUTGOING_PARSED_MSG_BATCH_QUANTITY = Counter('th2_mq_outgoing_parsed_msg_batch_quantity',
                                                 'Quantity of outgoing parsed message batches')
    OUTGOING_PARSED_MSG_QUANTITY = Counter('th2_mq_outgoing_parsed_msg_quantity',
                                           'Quantity of outgoing parsed messages')

    def get_delivery_counter(self) -> Counter:
        return self.OUTGOING_PARSED_MSG_BATCH_QUANTITY

    def get_content_counter(self) -> Counter:
        return self.OUTGOING_PARSED_MSG_QUANTITY

    def extract_count_from(self, message: MessageBatch):
        return len(self.get_messages(message))

    def get_messages(self, batch: MessageBatch) -> list:
        return batch.messages

    def value_to_bytes(self, value: MessageBatch):
        return value.SerializeToString()



#   Copyright 2020-2021 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABC, abstractmethod
from threading import Lock

from th2_common.schema.message.configuration.queue_configuration import QueueConfiguration
from th2_common.schema.message.impl.rabbitmq.connection.connection_manager import ConnectionManager
from th2_common.schema.message.message_sender import MessageSender
from th2_common.schema.message.message_subscriber import MessageSubscriber


class MessageQueue(ABC):

    def __init__(self, connection_manager: ConnectionManager, queue_configuration: QueueConfiguration) -> None:
        self.connection_manager = connection_manager
        self.queue_configuration = queue_configuration
        self.subscriber_lock = Lock()
        self.sender_lock = Lock()

    @abstractmethod
    def get_subscriber(self) -> MessageSubscriber:
        pass

    @abstractmethod
    def get_sender(self) -> MessageSender:
        pass

    @abstractmethod
    def close(self):
        pass



#   Copyright 2020-2020 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


from th2_common.schema.message.configuration.configuration import Configuration
from th2_common.schema.message.configuration.mq_router_filter_configuration import MqRouterFilterConfiguration


class QueueConfiguration(Configuration):

    def __init__(self, name: str, queue: str, exchange: str, attributes: list, filters: list,
                 prefetch_count: int = 1000, can_read=True, can_write=True) -> None:
        self.name = name
        self.queue = queue
        self.exchange = exchange
        self.prefetch_count = prefetch_count
        self.attributes = attributes
        self.filters = [MqRouterFilterConfiguration(**filter_schema) for filter_schema in filters]
        self.can_read = can_read
        self.can_write = can_write



#   Copyright 2020-2020 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


from typing import List

from google.protobuf.message import Message

from th2_common.schema.filter.strategy.filter_strategy import FilterStrategy
from th2_common.schema.message.configuration.field_filter_configuration import FieldFilterConfiguration
from th2_common.schema.message.configuration.router_filter import RouterFilter
from th2_common.schema.strategy.field_extraction.impl.th2_batch_msg_field_extraction import Th2BatchMsgFieldExtraction


class DefaultFilterStrategy(FilterStrategy):

    def __init__(self, extract_strategy=Th2BatchMsgFieldExtraction()) -> None:
        self.extract_strategy = extract_strategy

    def verify(self, message: Message, router_filter: RouterFilter = None, router_filters: List[RouterFilter] = None):
        if router_filters is None:
            msg_field_filters = dict(router_filter.get_message())
            msg_field_filters.update(router_filter.get_metadata())
            return self.check_values(self.extract_strategy.get_fields(message), msg_field_filters)
        else:
            if len(router_filters) == 0:
                return True
            for fields_filter in router_filters:
                if self.verify(message=message, router_filter=fields_filter):
                    return True
            return False

    def check_values(self, message_fields: {str: str}, field_filters: {str: FieldFilterConfiguration}) -> bool:
        for field_name in field_filters.keys():
            field_filter = field_filters[field_name]
            msg_field_value = message_fields[field_name]
            if not self.check_value(msg_field_value, field_filter):
                return False
        return True

    def check_value(self, value1, filter_configuration: FieldFilterConfiguration):
        if len(value1) == 0:
            return False

        value2 = filter_configuration.value
        if filter_configuration.operation == "EQUAL":
            return value1 == value2
        elif filter_configuration.operation == "NOT_EQUAL":
            return value1 != value2
        elif filter_configuration.operation == "EMPTY":
            return len(value1) == 0
        elif filter_configuration.operation == "NOT_EMPTY":
            return len(value1) != 0
        else:
            return False



#   Copyright 2020-2020 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


from importlib import import_module
from pathlib import Path
from pkgutil import iter_modules

import grpc

from th2_common.schema.exception.grpc_router_error import GrpcRouterError
from th2_common.schema.grpc.configuration.grpc_router_configuration import GrpcRouterConfiguration
from th2_common.schema.grpc.router.abstract_grpc_router import AbstractGrpcRouter
import th2_common.schema.strategy.route.impl as route


class DefaultGrpcRouter(AbstractGrpcRouter):

    def __init__(self, configuration: GrpcRouterConfiguration) -> None:
        super().__init__(configuration)
        self.strategies = dict()
        self.__load_strategies()

    def get_service(self, cls):
        return cls(self)

    class Connection:

        def __init__(self, service, strategy_obj, stub_class, channels):
            self.service = service
            self.strategy_obj = strategy_obj
            self.stubClass = stub_class
            self.channels = channels
            self.stubs = {}

        def __create_stub_if_not_exists(self, endpoint_name, config):
            socket = f"{config['host']}:{config['port']}"
            if socket not in self.channels:
                self.channels[socket] = grpc.insecure_channel(socket)

            if endpoint_name not in self.stubs:
                self.stubs[endpoint_name] = self.stubClass(self.channels[socket])

        def create_request(self, request_name, request, timeout):
            endpoint = self.strategy_obj.get_endpoint(request)
            endpoint_config = self.service['endpoints'][endpoint]
            if endpoint_config is not None:
                self.__create_stub_if_not_exists(endpoint, endpoint_config)
            stub = self.stubs[endpoint]
            if stub is not None:
                return getattr(stub, request_name)(request, timeout=timeout)

    def get_connection(self, service_class, stub_class):
        find_service = None
        if self.configuration.services:
            for service in self.configuration.services:
                if self.configuration.services[service]['service-class'].split('.')[-1] == service_class.__name__:
                    find_service = self.configuration.services[service]
                    break
        else:
            raise GrpcRouterError("Services list are empty in 'grpc.json'. Check your links")

        strategy_name = find_service['strategy']['name']
        strategy_class = self.strategies[strategy_name]
        if strategy_class is None:
            return None
        strategy_obj = strategy_class(find_service['strategy'])
        return self.Connection(find_service, strategy_obj, stub_class, self.channels)

    def __load_strategies(self):
        package_dir = Path(route.__file__).resolve().parent

        for _, module_name, _ in iter_modules([package_dir]):
            module = import_module(f'{route.__name__}.{module_name}')
            for name in dir(module):
                if not name.startswith('__'):
                    attr = getattr(module, name)
                    if dir(attr).__contains__('get_endpoint'):
                        self.strategies[name.lower()] = attr

        self.strategies.pop('routingstrategy', None)



#   Copyright 2020-2020 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


from google.protobuf.message import Message

from th2_common.schema.filter.strategy.impl.default_filter_strategy import DefaultFilterStrategy
from th2_common.schema.grpc.configuration.grpc_raw_filter_strategy import GrpcRawFilterStrategy
from th2_common.schema.strategy.route.routing_strategy import RoutingStrategy


class FilterRoutingStrategy(RoutingStrategy):

    def __init__(self, configuration: GrpcRawFilterStrategy) -> None:
        self.__filter_strategy = DefaultFilterStrategy()
        self.__grpc_configuration = configuration

    def get_endpoint(self, message: Message):
        endpoint: set = self.__filter(message)
        if len(endpoint) != 1:
            raise Exception('Wrong size of endpoints for send. Should be equal to 1')
        return endpoint.__iter__().__next__()

    def __filter(self, message: Message) -> {str}:
        endpoints = set()
        for fields_filter in self.__grpc_configuration.filters:
            if len(fields_filter.message) == 0 and len(fields_filter.metadata) == 0 \
                    or self.__filter_strategy.verify(message=message, router_filter=fields_filter):
                endpoints.add(fields_filter.endpoint)
        return endpoints



#   Copyright 2020-2021 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABC, abstractmethod

from th2_common.schema.exception.router_error import RouterError
from th2_common.schema.message.configuration.queue_configuration import QueueConfiguration
from th2_common.schema.message.impl.rabbitmq.connection.connection_manager import ConnectionManager
from th2_common.schema.message.message_queue import MessageQueue
from th2_common.schema.message.message_sender import MessageSender
from th2_common.schema.message.message_subscriber import MessageSubscriber


class AbstractRabbitQueue(MessageQueue, ABC):

    def __init__(self, connection_manager: ConnectionManager, queue_configuration: QueueConfiguration) -> None:
        super().__init__(connection_manager, queue_configuration)
        self.subscriber = None
        self.sender = None

    def get_subscriber(self) -> MessageSubscriber:
        if self.connection_manager is None or self.queue_configuration is None:
            raise RouterError('Queue not yet init')
        if not self.queue_configuration.can_read:
            raise RouterError('Queue can not read')
        with self.subscriber_lock:
            if self.subscriber is None or self.subscriber.is_close():
                self.subscriber = self.create_subscriber(self.connection_manager, self.queue_configuration)
            return self.subscriber

    def get_sender(self) -> MessageSender:
        if self.connection_manager is None or self.queue_configuration is None:
            raise RouterError('Queue not yet init')
        if not self.queue_configuration.can_write:
            raise RouterError('Queue can not write')
        with self.sender_lock:
            if self.sender is None or self.sender.is_close():
                self.sender = self.create_sender(self.connection_manager, self.queue_configuration)
            return self.sender

    def close(self):
        with self.subscriber_lock:
            if self.subscriber is not None and not self.subscriber.is_close():
                self.subscriber.close()
        with self.sender_lock:
            if self.sender is not None and not self.sender.is_close():
                self.sender.close()

    @abstractmethod
    def create_sender(self, connection_manager: ConnectionManager,
                      queue_configuration: QueueConfiguration) -> MessageSender:
        pass

    @abstractmethod
    def create_subscriber(self, connection_manager: ConnectionManager,
                          queue_configuration: QueueConfiguration) -> MessageSubscriber:
        pass



#   Copyright 2020-2021 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABC, abstractmethod
from threading import Lock

from th2_common.schema.exception.router_error import RouterError
from th2_common.schema.filter.strategy.filter_strategy import FilterStrategy
from th2_common.schema.filter.strategy.impl.default_filter_strategy import DefaultFilterStrategy
from th2_common.schema.message.configuration.queue_configuration import QueueConfiguration
from th2_common.schema.message.impl.rabbitmq.connection.connection_manager import ConnectionManager
from th2_common.schema.message.message_listener import MessageListener
from th2_common.schema.message.message_queue import MessageQueue
from th2_common.schema.message.message_router import MessageRouter
from th2_common.schema.message.message_subscriber import MessageSubscriber
from th2_common.schema.message.subscriber_monitor import SubscriberMonitor


class SubscriberMonitorImpl(SubscriberMonitor):

    def __init__(self, subscriber: MessageSubscriber, lock=Lock()) -> None:
        self.lock = lock
        self.subscriber = subscriber

    def unsubscribe(self):
        with self.lock:
            self.subscriber.close()


class MultiplySubscribeMonitorImpl(SubscriberMonitor):

    def __init__(self, subscriber_monitors: list) -> None:
        self.subscriber_monitors = subscriber_monitors

    def unsubscribe(self):
        for monitor in self.subscriber_monitors:
            monitor.unsubscribe()


class AbstractRabbitMessageRouter(MessageRouter, ABC):

    def __init__(self, connection_manager, configuration) -> None:
        super().__init__(connection_manager, configuration)
        self._filter_strategy = DefaultFilterStrategy()
        self.queue_connections = dict()
        self.queue_connections_lock = Lock()

    @property
    @abstractmethod
    def required_subscribe_attributes(self):
        pass

    @property
    @abstractmethod
    def required_send_attributes(self):
        pass

    def add_subscribe_attributes(self, queue_attr):
        return self.required_subscribe_attributes.union(queue_attr)

    def add_send_attributes(self, queue_attr):
        return self.required_send_attributes.union(queue_attr)

    def _subscribe_by_alias(self, callback: MessageListener, queue_alias) -> SubscriberMonitor:
        queue: MessageQueue = self._get_message_queue(queue_alias)
        subscriber: MessageSubscriber = queue.get_subscriber()
        subscriber.add_listener(callback)
        try:
            subscriber.start()
        except Exception as e:
            raise RuntimeError('Can not start subscriber', e)
        return SubscriberMonitorImpl(subscriber, queue.subscriber_lock)

    def subscribe(self, callback: MessageListener, *queue_attr) -> SubscriberMonitor:
        attrs = self.add_subscribe_attributes(queue_attr)
        queues = self.configuration.find_queues_by_attr(attrs)
        if len(queues) != 1:
            raise RouterError(
                f'Wrong amount of queues for subscribe. '
                f'Found {len(queues)} queues, but must be only 1. Search was done by {queue_attr} attributes')
        return self._subscribe_by_alias(callback, queues.keys().__iter__().__next__())

    def subscribe_all(self, callback: MessageListener, *queue_attr) -> SubscriberMonitor:
        attrs = self.add_subscribe_attributes(queue_attr)
        subscribers = []
        for queue_alias in self.configuration.find_queues_by_attr(attrs).keys():
            subscribers.append(self._subscribe_by_alias(callback, queue_alias))
        if len(subscribers) == 0:
            raise RouterError(f'Wrong amount of queues for subscribe_all_by_attr. Must not be empty. '
                              f'Search was done by {attrs} attributes')
        return MultiplySubscribeMonitorImpl(subscribers)

    def unsubscribe_all(self):
        with self.queue_connections_lock:
            for queue in self.queue_connections.values():
                queue.close()
            self.queue_connections.clear()

    def close(self):
        self.unsubscribe_all()

    def send(self, message, *queue_attr):
        if not queue_attr:
            self._send_by_aliases_and_messages_to_send(self._find_by_filter(self.configuration.queues, message))
            return

        attrs = self.add_send_attributes(queue_attr)
        filtered_by_attr = self.configuration.find_queues_by_attr(attrs)
        filtered_by_attr_and_filter = self._find_by_filter(filtered_by_attr, message)
        if len(filtered_by_attr_and_filter) != 1:
            raise Exception(f'Wrong amount of queues for send. Must be equal to 1. '
                            f'Found {len(filtered_by_attr_and_filter)} queues, but must be only 1. '
                            f'Search was done by {attrs} attributes')
        self._send_by_aliases_and_messages_to_send(filtered_by_attr_and_filter)

    def send_all(self, message, *queue_attr):
        attrs = self.add_send_attributes(queue_attr)
        filtered_by_attr = self.configuration.find_queues_by_attr(attrs)
        filtered_by_attr_and_filter = self._find_by_filter(filtered_by_attr, message)
        if len(filtered_by_attr_and_filter) == 0:
            raise Exception(f'Wrong amount of queues for send_all. Must not be equal to 0. '
                            f'Search was done by {attrs} attributes')
        self._send_by_aliases_and_messages_to_send(filtered_by_attr_and_filter)

    def set_filter_strategy(self, filter_strategy: FilterStrategy):
        self._filter_strategy = filter_strategy

    @abstractmethod
    def _create_queue(self, connection_manager: ConnectionManager,
                      queue_configuration: QueueConfiguration) -> MessageQueue:
        pass

    @abstractmethod
    def _find_by_filter(self, queues: {str: QueueConfiguration}, msg) -> dict:
        pass

    def _send_by_aliases_and_messages_to_send(self, aliases_and_messages_to_send: dict):
        for queue_alias, message in aliases_and_messages_to_send.items():
            try:
                sender = self._get_message_queue(queue_alias).get_sender()
                sender.start()
                sender.send(message)
            except Exception:
                raise RouterError('Can not start sender')

    def _get_message_queue(self, queue_alias) -> MessageQueue:
        with self.queue_connections_lock:
            if queue_alias not in self.queue_connections:
                self.queue_connections[queue_alias] = self._create_queue(self.connection_manager,
                                                                         self.configuration.get_queue_by_alias(
                                                                             queue_alias))
            return self.queue_connections[queue_alias]



#   Copyright 2020-2020 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


from abc import ABC, abstractmethod

import google.protobuf.message
from th2_grpc_common.common_pb2 import Direction, Message

from th2_common.schema.strategy.field_extraction.field_extraction_strategy import FieldExtractionStrategy


class AbstractTh2MsgFieldExtraction(FieldExtractionStrategy, ABC):

    SESSION_ALIAS_KEY = 'session_alias'
    MESSAGE_TYPE_KEY = 'message_type'
    DIRECTION_KEY = 'direction'

    def get_fields(self, message: google.protobuf.message.Message) -> {str: str}:
        th2msg = self.parse_message(message)
        message_id = th2msg.metadata.id

        message_fields = dict()
        for field_name in th2msg.fields.keys():
            message_fields[field_name] = th2msg.fields[field_name].simple_value
        metadata_msg_fields = {self.SESSION_ALIAS_KEY: message_id.connection_id.session_alias,
                               self.MESSAGE_TYPE_KEY: th2msg.metadata.message_type,
                               self.DIRECTION_KEY: Direction.Name(message_id.direction)}
        message_fields.update(metadata_msg_fields)
        return message_fields

    @abstractmethod
    def parse_message(self, message: google.protobuf.message.Message) -> Message:
        pass



#   Copyright 2020-2021 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from prometheus_client.exposition import start_wsgi_server

from th2_common.schema.metrics.common_metrics import CommonMetrics


class PrometheusServer:

    def __init__(self, port=8000, host='localhost'):
        self.stopped = None
        self.port = port
        self.host = host

    def run(self):
        CommonMetrics.LIVENESS.inc()
        start_wsgi_server(self.port, self.host)
        self.stopped = False

    def stop(self):
        CommonMetrics.LIVENESS.dec()
        self.stopped = True



#   Copyright 2020-2021 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from th2_common.schema.event.event_batch_queue import EventBatchQueue
from th2_common.schema.message.configuration.queue_configuration import QueueConfiguration
from th2_common.schema.message.impl.rabbitmq.abstract_rabbit_message_router import AbstractRabbitMessageRouter
from th2_common.schema.message.impl.rabbitmq.connection.connection_manager import ConnectionManager
from th2_common.schema.message.message_queue import MessageQueue
from th2_common.schema.message.queue_attribute import QueueAttribute


class EventBatchRouter(AbstractRabbitMessageRouter):

    @property
    def required_subscribe_attributes(self):
        return {QueueAttribute.SUBSCRIBE.value, QueueAttribute.EVENT.value}

    @property
    def required_send_attributes(self):
        return {QueueAttribute.PUBLISH.value, QueueAttribute.EVENT.value}

    def _create_queue(self, connection_manager: ConnectionManager,
                      queue_configuration: QueueConfiguration) -> MessageQueue:
        return EventBatchQueue(connection_manager, queue_configuration)

    def _find_by_filter(self, queues: {str: QueueConfiguration}, msg) -> dict:
        return {key: msg for key in queues.keys()}



import logging
import threading
import time

import pika

from th2_common.schema.message.impl.rabbitmq.configuration.rabbitmq_configuration import RabbitMQConfiguration

logger = logging.getLogger()


class ConnectionManager:

    def __init__(self, config: RabbitMQConfiguration) -> None:
        self.configuration: RabbitMQConfiguration = config
        self.__credentials = pika.PlainCredentials(config.username,
                                                   config.password)
        self.__connection_parameters = pika.ConnectionParameters(virtual_host=config.vhost,
                                                                 host=config.host,
                                                                 port=config.port,
                                                                 credentials=self.__credentials)
        self.connection = None
        self.__connection_thread = None
        self.connection_lock = threading.Lock()
        self.open_connection()

        self.reconnect_attempts = 10

    def open_connection(self):
        with self.connection_lock:
            connection_open_timeout = 60
            self.connection = pika.SelectConnection(self.__connection_parameters)
            self.__connection_thread = threading.Thread(target=self.__run_connection_thread)
            self.__connection_thread.start()
            for x in range(int(connection_open_timeout / 5)):
                if not self.connection.is_open:
                    time.sleep(5)
            if not self.connection.is_open:
                raise ConnectionError(f'The connection has not been opened for {connection_open_timeout} seconds')
            logger.info(f'Connection is open')

    def close_connection(self):
        with self.connection_lock:
            connection_close_timeout = 60
            self.connection.close()
            self.__connection_thread.join(connection_close_timeout)
            logger.info(f'Connection is close')

    def reopen_connection(self):
        self.close_connection()
        self.open_connection()

    def __run_connection_thread(self):
        try:
            self.connection.ioloop.start()
        except Exception:
            logger.exception(f'Failed starting loop SelectConnection')

    def close(self):
        self.close_connection()



#   Copyright 2020-2020 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


import json
import logging
import os
from abc import ABC, abstractmethod
from threading import Lock

from th2_common.schema.cradle.cradle_configuration import CradleConfiguration
from th2_common.schema.event.event_batch_router import EventBatchRouter
from th2_common.schema.grpc.configuration.grpc_router_configuration import GrpcRouterConfiguration
from th2_common.schema.grpc.router.grpc_router import GrpcRouter
from th2_common.schema.grpc.router.impl.default_grpc_router import DefaultGrpcRouter
from th2_common.schema.message.configuration.message_router_configuration import MessageRouterConfiguration
from th2_common.schema.message.impl.rabbitmq.configuration.rabbitmq_configuration import RabbitMQConfiguration
from th2_common.schema.message.impl.rabbitmq.connection.connection_manager import ConnectionManager
from th2_common.schema.message.impl.rabbitmq.parsed.rabbit_parsed_batch_router import RabbitParsedBatchRouter
from th2_common.schema.message.impl.rabbitmq.raw.rabbit_raw_batch_router import RabbitRawBatchRouter
from th2_common.schema.message.message_router import MessageRouter
from th2_common.schema.metrics.prometheus_configuration import PrometheusConfiguration
from th2_common.schema.metrics.prometheus_server import PrometheusServer

logger = logging.getLogger()


class AbstractCommonFactory(ABC):

    def __init__(self,
                 message_parsed_batch_router_class=RabbitParsedBatchRouter,
                 message_raw_batch_router_class=RabbitRawBatchRouter,
                 event_batch_router_class=EventBatchRouter,
                 grpc_router_class=DefaultGrpcRouter) -> None:

        self.rabbit_mq_configuration = self._create_rabbit_mq_configuration()
        self.message_router_configuration = self._create_message_router_configuration()
        self.grpc_router_configuration = None

        self.message_parsed_batch_router_class = message_parsed_batch_router_class
        self.message_raw_batch_router_class = message_raw_batch_router_class
        self.event_batch_router_class = event_batch_router_class
        self.grpc_router_class = grpc_router_class

        self._message_parsed_batch_router = None
        self._message_raw_batch_router = None
        self._event_batch_router = None
        self._grpc_router = None

        self._connection_manager = ConnectionManager(self.rabbit_mq_configuration)

        self.prometheus_config = PrometheusConfiguration()
        self.prometheus = PrometheusServer(self.prometheus_config.port, self.prometheus_config.host)
        if self.prometheus_config.enabled is True:
            self.prometheus.run()

    @property
    def message_parsed_batch_router(self) -> MessageRouter:
        """
        Created MessageRouter which work with MessageBatch
        """
        if self._message_parsed_batch_router is None:
            self._message_parsed_batch_router = self.message_parsed_batch_router_class(self._connection_manager,
                                                                                       self.message_router_configuration
                                                                                       )

        return self._message_parsed_batch_router

    @property
    def message_raw_batch_router(self) -> MessageRouter:
        """
        Created MessageRouter which work with RawMessageBatch
        """
        if self._message_raw_batch_router is None:
            self._message_raw_batch_router = self.message_raw_batch_router_class(self._connection_manager,
                                                                                 self.message_router_configuration)
        return self._message_raw_batch_router

    @property
    def event_batch_router(self) -> MessageRouter:
        """
        Created MessageRouter which work with EventBatch
        """
        if self._event_batch_router is None:
            self._event_batch_router = self.event_batch_router_class(self._connection_manager,
                                                                     self.message_router_configuration)

        return self._event_batch_router

    @property
    def grpc_router(self) -> GrpcRouter:
        if self._grpc_router is None:
            if self.grpc_router_configuration is None:
                self.grpc_router_configuration = self._create_grpc_router_configuration()
            self._grpc_router = self.grpc_router_class(self.grpc_router_configuration)

        return self._grpc_router

    def close(self):
        logger.info('Closing Common Factory')

        if self._message_raw_batch_router is not None:
            try:
                self._message_raw_batch_router.close()
            except Exception:
                logger.exception('Error during closing Message Router (Message Raw Batch)')

        if self._message_parsed_batch_router is not None:
            try:
                self._message_parsed_batch_router.close()
            except Exception:
                logger.exception('Error during closing Message Router (Message Parsed Batch)')

        if self._event_batch_router is not None:
            try:
                self._event_batch_router.close()
            except Exception:
                logger.exception('Error during closing Message Router (Event Batch)')

        if self._grpc_router is not None:
            try:
                self._grpc_router.close()
            except Exception:
                logger.exception('Error during closing gRPC Router')

        if self._connection_manager is not None:
            self._connection_manager.close()

        if self.prometheus.stopped is False:
            self.prometheus.stop()

    @staticmethod
    def read_configuration(filepath):
        with open(filepath, 'r') as file:
            config_json = file.read()
            config_json_expanded = os.path.expandvars(config_json)
            config_dict = json.loads(config_json_expanded)

        return config_dict

    def create_cradle_configuration(self) -> CradleConfiguration:
        return CradleConfiguration(**self.read_configuration(self._path_to_cradle_configuration()))

    def create_prometheus_configuration(self) -> PrometheusConfiguration:
        return PrometheusConfiguration(**self.read_configuration(self._path_to_prometheus_configuration()))

    def create_custom_configuration(self) -> dict:
        return self.read_configuration(self._path_to_custom_configuration())

    def _create_rabbit_mq_configuration(self) -> RabbitMQConfiguration:
        lock = Lock()
        try:
            lock.acquire()
            if not hasattr(self, 'rabbit_mq_configuration'):
                config_dict = self.read_configuration(self._path_to_rabbit_mq_configuration())
                self.rabbit_mq_configuration = RabbitMQConfiguration(**config_dict)
        finally:
            lock.release()
        return self.rabbit_mq_configuration

    def _create_message_router_configuration(self) -> MessageRouterConfiguration:
        lock = Lock()
        with lock:
            if not hasattr(self, 'message_router_configuration'):
                config_dict = self.read_configuration(self._path_to_message_router_configuration())
                self.message_router_configuration = MessageRouterConfiguration(**config_dict)
        return self.message_router_configuration

    def _create_grpc_router_configuration(self) -> GrpcRouterConfiguration:
        lock = Lock()
        try:
            lock.acquire()
            config_dict = self.read_configuration(self._path_to_grpc_router_configuration())
            self.grpc_router_configuration = GrpcRouterConfiguration(**config_dict)
        finally:
            lock.release()
        return self.grpc_router_configuration

    @abstractmethod
    def _path_to_rabbit_mq_configuration(self) -> str:
        pass

    @abstractmethod
    def _path_to_message_router_configuration(self) -> str:
        pass

    @abstractmethod
    def _path_to_grpc_router_configuration(self) -> str:
        pass

    @abstractmethod
    def _path_to_cradle_configuration(self) -> str:
        pass

    @abstractmethod
    def _path_to_prometheus_configuration(self) -> str:
        pass

    @abstractmethod
    def _path_to_custom_configuration(self) -> str:
        pass



#   Copyright 2020-2021 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from th2_grpc_common.common_pb2 import RawMessageBatch, RawMessage

from th2_common.schema.message.configuration.queue_configuration import QueueConfiguration
from th2_common.schema.message.impl.rabbitmq.connection.connection_manager import ConnectionManager
from th2_common.schema.message.impl.rabbitmq.raw.rabbit_raw_batch_queue import RabbitRawBatchQueue
from th2_common.schema.message.impl.rabbitmq.router.abstract_rabbit_batch_message_router import \
    AbstractRabbitBatchMessageRouter
from th2_common.schema.message.message_queue import MessageQueue
from th2_common.schema.message.queue_attribute import QueueAttribute


class RabbitRawBatchRouter(AbstractRabbitBatchMessageRouter):

    @property
    def required_subscribe_attributes(self):
        return {QueueAttribute.SUBSCRIBE.value, QueueAttribute.RAW.value}

    @property
    def required_send_attributes(self):
        return {QueueAttribute.PUBLISH.value, QueueAttribute.RAW.value}

    def _get_messages(self, batch: RawMessageBatch) -> list:
        return batch.messages

    def _create_batch(self):
        return RawMessageBatch()

    def _add_message(self, batch: RawMessageBatch, message: RawMessage):
        batch.messages.append(message)

    def _create_queue(self, connection_manager: ConnectionManager,
                      queue_configuration: QueueConfiguration) -> MessageQueue:
        return RabbitRawBatchQueue(connection_manager, queue_configuration)
